{$ifdef CPU386}
{$define GCDOpt}
{$endif}
unit Rational;
{
Operations with fractional numbers
for Delphi 3.0
Copyright (c) 1996-1997 Dmitriy Anisimkov
e-mail anisimkov@yahoo.com
phone +7 (381)-2217774
}
interface

uses math;

type

tRational = record
  numer : Comp;
  denom : Comp;
end;

TElement = TRational;

function CreateRational(const num,denom : Comp) : tRational;
function Compare(const r1,r2 : tRational) : Extended;
function IntPower(const r : tRational;Exponent : integer) : tRational;
function Inverse(const r : tRational) : tRational;
function MulRC(const r : tRational;c : Comp) : tRational;
function Multiply(const r1,r2 : tRational) : tRational;
function Negative(const r : tRational) : tRational;
function Fract(const r : tRational) : tRational;
function PositiveFract(const r : tRational) : tRational;
function RationalToStr(r : tRational) : String;
function Sum(const r1,r2 : tRational) : tRational;
function SumRC(const r : tRational;c : Comp) : tRational;
function ToDouble(r : tRational) : Double;
function Trunc(const r : tRational) : TRational;

function Euclide(const a,b : comp;var x,y : Comp) : Comp;
function GreatestDivisor(const c1,c2 : comp) : comp;
function LowestFactor(const c1,c2 : comp) : comp;
function MinMembBetw(const r1,r2 : tRational) : tRational;
function ModComp(c1,c2 : comp) : comp;
function Represent(const c1,c2 : comp) : comp;
function DoubleToRational(d,pres : Double) : tRational;

const
  NegativeOne : tRational = (numer : -1;denom : 1);
  One : tRational = (numer : 1;denom : 1);
  NULL : tRational = (numer : 0;denom : 1);
  UNLIMITED : tRational = (numer : 1;denom : 0);
  NegativeUNLIMITED : tRational = (numer : -1;denom : 0);
  Undefined : tRational = (numer : 0;denom : 0);

implementation

uses SysUtils
//,utils
;

function ModComp(c1,c2 : comp) : comp;
asm
  FILD c2;
  FILD c1;
  FPREM;
  FSTP ST(1);
  FWAIT;
end;

function Trunc(const r : tRational) : TRational;
begin
  result.numer := r.numer/r.denom;
  result.denom := 1;
end;

function Euclide(const a,b : comp;var x,y : Comp) : Comp;
begin
  if b=0 then begin
    result := a;
    x := 1;
    y := 1;
  end else begin
    result := euclide(b,ModComp(a,b),y,x);
    x := y*Int(a/b)-1;
  end;
end;
//    raise Exception.Create('not implemented');

{$ifdef CheckND}
procedure ChkND(const r : tRational);
begin
  if r.denom<0 then
    raise Exception.Create('Negative denominator');
end;
{$endif}

function LowestFactor(const c1,c2 : comp) : comp;
begin
  result := c1/GreatestDivisor(c1,c2)*c2;
end;

function GreatestDivisor(const c1,c2 : comp) : comp;
{$ifndef GCDOpt}
begin
  if c2=0 then result := Abs(c1)
  else result := GreatestDivisor(c2,ModComp(c1,c2));
{$else}
asm
  FILD c1
  FILD c2
@@1:
  FTST
  FSTSW AX
  test ah,40h
  jnz @@2
  FXCH
  FPREM
  jmp @@1
@@2:
  FFREE ST(0)
  FINCSTP;
  FABS
  FWAIT
{$endif}
end;

function MinMembBetw(const r1,r2 : tRational) : tRational;
// r1>=0,r2>=0,r1<{=?}r2
var
  c,c2 : comp;
begin
  c := Int(r1.numer/r1.denom);
  c2 := Int(r2.numer/r2.denom);
  if c<c2 then begin
    result.numer := c+1;
    result.denom := 1;
  end else
    result := SumRC(Inverse(MinMembBetw(
    Inverse(SumRC(r2,-c)),Inverse(SumRC(r1,-c))
    )),c);
end;

function Represent(const c1,c2 : comp) : comp;
begin
  result := ModComp(c1,c2);
  if result<0 then result := c2+result;
end;

function PositiveFract(const r : tRational) : tRational;
begin
  result.numer := Represent(r.numer,r.denom);
  result.denom := r.denom;
end;

function Fract(const r : tRational) : tRational;
begin
  result.numer := ModComp(r.numer,r.denom);
  result.denom := r.denom;
end;

function DoubleToRational(d,pres : Double) : tRational;
begin
  result := MinMembBetw(CreateRational(d/pres-1,1/pres),CreateRational(d/pres+1,1/pres));
end;

function ToDouble(r : tRational) : Double;
begin
  result := r.numer/r.denom;
end;

procedure ReduceComps(var c1,c2 : Comp);
var
  cdev : comp;
begin
  cdev := GreatestDivisor(c1,c2);
  c1 := c1/cdev;
  c2 := c2/cdev;
end;

function Reduced(const r : tRational) : tRational;
var
  cdev : comp;
begin
  cdev := GreatestDivisor(r.numer,r.denom);
  result.numer := r.numer/cdev;
  result.denom := r.denom/cdev;
end;

function CreateRational(const num,denom : Comp) : tRational;
var
  cdev : comp;
begin
  cdev := GreatestDivisor(num,denom);
  if denom>0 then begin
    result.Numer := num/cdev;
    result.Denom := denom/cdev;
  end else begin
    result.Numer := -num/cdev;
    result.Denom := -denom/cdev;
  end;
{$ifdef CheckND}
  ChkND(result);
{$endif}
end;

function Compare(const r1,r2 : tRational) : Extended;
begin
  result := r1.numer*r2.denom-r1.denom*r2.numer;
  if (result=0) and not CompareMem(@r1,@r2,sizeof(r1)) then
    result := Sum(r1,negative(r2)).numer;
end;

function RationalToStr(r : tRational) : String;
begin
  ReduceComps(r.numer,r.denom);
  result := floatToStr(r.numer);
  if r.denom<>1.0 then appendStr(result,'/'+floatToStr(r.denom));
end;

function IntPower(const r : tRational;Exponent : integer) : tRational;
var
  cn : integer;
begin
  result := One;
  cn := Abs(Exponent);
  while cn>0 do begin
    result := Multiply(result,r);
    dec(cn);
  end;
  if Exponent<0 then result := Inverse(result);
{$ifdef CheckND}
  ChkND(result);
{$endif}
end;

function SumRC(const r : tRational;c : Comp) : tRational;
begin
  result.denom := r.denom;
  result.numer := c*r.denom;
  result.numer := result.numer+r.numer;
end;

function Sum(const r1,r2 : tRational) : tRational;
var
  nd : tRational;
  c1,c2 : comp;
begin
  if r1.numer=0 then begin
    result := r2;
    exit;
  end;
  if r2.numer=0 then begin
    result := r1;
    exit;
  end;
  nd := CreateRational(GreatestDivisor(r1.numer,r2.numer),GreatestDivisor(r1.denom,r2.denom));
  c1 := (r1.numer/nd.numer)*(r2.denom/nd.denom);
  c2 := (r1.denom/nd.denom)*(r2.numer/nd.numer);
  result.numer := c1+c2;
  {c1 and c2 for check losing explicit}
  result.denom := (r1.denom/nd.denom)*(r2.denom/nd.denom);
  result := Multiply(Reduced(result),nd);
{$ifdef CheckND}
  ChkND(result);
{$endif}
end;

function Multiply(const r1,r2 : tRational) : tRational;
var
  rt1 : tRational;
  rt2 : tRational;
begin
  rt1 := CreateRational(r1.numer,r2.denom);
  rt2 := CreateRational(r2.numer,r1.denom);
  result.numer := rt1.numer*rt2.numer;
  result.denom := rt1.denom*rt2.denom;
{$ifdef CheckND}
  ChkND(result);
{$endif}
end;

function MulRC(const r : tRational;c : Comp) : tRational;
begin
  Result := CreateRational(r.numer*c,r.denom);
end;

function Inverse(const r : tRational) : tRational;
begin
  if r.numer<0 then begin
    result.numer := -r.denom;
    result.denom := -r.numer;
  end else begin
    result.numer := r.denom;
    result.denom := r.numer;
  end;
{$ifdef CheckND}
  ChkND(result);
{$endif}
end;

function Negative(const r : tRational) : tRational;
begin
  result.numer := -r.numer;
  result.denom := r.denom;
{$ifdef CheckND}
  ChkND(result);
{$endif}
end;

end.
