unit TestForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, Spin, LPRat, Menus,
  Gauges,  ExtCtrls,MxEl;

type
pArrayOfDouble = ^tArrayOfDouble;
tArrayOfDouble = array [0..MaxInt div sizeof(double)-1] of double;
pArrayOfRect = ^TArrayOfRect;
tArrayOfRect = array [0..MaxInt div sizeof(tRect)-1] of tRect;

  TtstLPForm = class(TForm)
    MainMenu: TMainMenu;
    FileSubMenu: TMenuItem;
    Open: TMenuItem;
    Save: TMenuItem;
    Saveas: TMenuItem;
    SaveDialog: TSaveDialog;
    OpenDialog: TOpenDialog;
    StatePanel: TPanel;
    LPProgress: TGauge;
    DurationLab: TLabel;
    CastLab: TLabel;
    Dimlab: TLabel;
    Exponent: TLabel;
    ExpEd: TSpinEdit;
    Solution: TButton;
    Clear: TButton;
    CoordLab: TLabel;
    MinLines: TButton;
    MaxBox: TCheckBox;
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ClearClick(Sender: TObject);
    procedure SolutionClick(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormPaint(Sender: TObject);
    procedure SaveasClick(Sender: TObject);
    procedure OpenClick(Sender: TObject);
    procedure SaveClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure MinLinesClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
    LastX,LastY : integer;
    m : tLPMatrix;
    res  : pArrayOfDouble;
    LineCount : integer;
    Lines : pArrayOfRect;
    StartTime : tDateTime;
    SkipTime : tDateTime;
    skip  : integer;
    shlist : tStringList;
    NodesCount : integer;
    FileName : String;
    function IsCancel : boolean;
    procedure DrawPolinom(const cf : array of Double);
    procedure OnShowSkip(sender : tLPMatrix);
    procedure OnChangeLeader(m : tLPMatrix);
    procedure OnSimplex(m : tLPMatrix);
    procedure OnString(sender : tLPMatrix);
    procedure SkipNode(sender : tLPMatrix);
    procedure DelPoint(Sender: TObject; Button: TMouseButton;
    Shift: TShiftState; X, Y: Integer);
    procedure ShMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    function SizeBetween(idx  : integer) : TElement;
//    procedure PreUBSolution;
  public
    procedure OnLPSkip(sender : tLPMatrix);
    { Public declarations }
  end;

var
  tstLPForm: TtstLPForm;

implementation

uses math,LPMatxF,Matrix,utils;
{$R *.DFM}

type
  tScrPoint = class(TBevel);

const
  vc = 1;
  DeltaBeep = 20/24/60/60;

function DurationToStr(DurT : TDateTime) : String;
var
  Hour,Min,Sec,MSec : Word;
begin
  DecodeTime(Frac(DurT),Hour,Min,Sec,MSec);
  result := IntToStr(system.trunc(DurT))+' '+TimeToStr(Frac(DurT))+'.'+Int2Str(MSec,3);
end;

procedure TtstLPForm.DelPoint(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button<>mbRight then exit;
  with ShList do Delete(IndexOfObject(Sender));
  Sender.Free;
  DimLab.Caption := IntToStr(ShList.Count);
  if ShList.Count=0 then begin
    Save.Enabled := false;
    SaveAs.Enabled := false;
  end;
end;

procedure TtstLPForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  shp : tScrPoint;
begin
  if Button<>mbLeft then exit;
  shp := tScrPoint.Create(Self);
  shp.OnMouseDown := DelPoint;
  shp.OnMouseMove := ShMouseMove;
  LastY := 1;
  LastX := 1;
  insertControl(shp);
  ShList.AddObject('',shp);
  DimLab.Caption := IntToStr(ShList.Count);
  Save.Enabled := true;
  SaveAs.Enabled := true;
  with shp do begin
    left := x;
    top := y;
    width := 10;
    height := 10;
  end;
end;

function TtstLPForm.IsCancel : boolean;
begin
  result := Solution.Enabled and MinLines.Enabled
  or (MessageDlg('Cancel the solution ?',mtWarning,[mbYes,mbNo],0)=mrYes);
end;

procedure TtstLPForm.ClearClick(Sender: TObject);
var
  cn : integer;
begin
  if not IsCancel then exit;
  cn := 0;
  Save.Enabled := false;
  SaveAs.Enabled := false;
  ShList.Clear;
  while cn<ComponentCount do begin
    if Components[cn] is TScrPoint then Components[cn].Free
    else inc(cn);
  end;
  FileName := '';
  reallocMem(res,0);
  reallocmem(lines,0);
  refresh;
end;

{function ClassList(cmp : tComponent;cl : tClass) : tStringList;
var
  cn : integer;
begin
  result := tStringList.Create;
  with cmp do
    for cn := 0 to ComponentCount-1 do
    if Components[cn] is Cl then result.AddObject('',Components[cn]);
end;}

procedure TtstLPForm.DrawPolinom(const cf : array of Double);
var
  r : tRect;
  cn,cp : integer;
  psum : Double;
  prevdraw : boolean;
begin
  r := Canvas.ClipRect;
  prevdraw := false;
  for cn := r.left to r.Right do begin
    psum := 0;
    for cp := 0 to high(cf) do psum := psum+cf[cp]*Power(cn/LastX,cp);
    psum := psum*LastY;
    if (psum>r.top) and (psum<r.Bottom) then begin
      if prevdraw then Canvas.lineto(cn,Round(psum))
      else Canvas.moveto(cn,Round(psum));
      prevdraw := true;
    end else prevdraw := false;
  end;
end;

procedure Pack2Arr(var pa : array of Double);
var
  cn : integer;
begin
  for cn := 0 to High(pa) div 2 do begin
    pa[cn] := Pa[2*cn+1]-pa[2*cn];
  end;
end;

function TtstLPForm.SizeBetween(idx  : integer) : TElement;
var
  c1,c2 : TControl;
begin
  c1 := ShList.Objects[idx div ShList.Count] as tControl;
  c2 := ShList.Objects[idx mod ShList.Count] as tControl;
  result.numer := sqrt(Sqr(c1.Left - c2.Left)+
  Sqr(c1.Top - c2.Top))*10*(ord(not MaxBox.Checked)*2-1);
  result.denom := 10;
end;

procedure TtstLPForm.OnSimplex(m : tLPMatrix);
begin
  application.processmessages;
  if ShList.Count=0 then raise EAbort.Create('');
end;

procedure TtstLPForm.SkipNode(sender : tLPMatrix);
var
  n : integer;
const
  bits = 30;
begin
  if not assigned(sender.copiedFrom) then exit;
  n := sender.copiedFrom.MaxVar-sender.MaxVar;
  if n>bits then exit;
  inc(NodesCount,1 shl (bits-n));
  CastLab.Caption := IntToHex(NodesCount,8);
end;

procedure TtstLPForm.OnChangeLeader(m : tLPMatrix);
var
  ci : integer;
  i,j,idx : integer;
  SumLen : TElement;
begin
  ci := 0;
  SumLen := null;
  for i := 0 to shList.Count-2 do begin
    for j := i+1 to shList.Count-1 do begin
      idx := i*shList.Count+j;
      if Compare(m.Variables[idx].value,One)=0 then begin
        if ci<LineCount then begin
        Lines[ci].TopLeft.x := tControl(ShList.Objects[idx div ShList.Count]).left;
        Lines[ci].TopLeft.y := tControl(ShList.Objects[idx div ShList.Count]).top;
        Lines[ci].BottomRight.x := tControl(ShList.Objects[idx mod ShList.Count]).left;
        Lines[ci].BottomRight.y := tControl(ShList.Objects[idx mod ShList.Count]).top;
        end;
        SumLen := mxel.Sum(SumLen,SizeBetween(idx));
        inc(ci);
      end else if m.Variables[idx].value.numer<>0 then
        ErrMsg('Var unposable');
    end;
  end;
  if Compare(SumLen,Negative(m[0,0]))<>0 then
    ErrMsg('Check Error');
  if ci<>LineCount then
    ErrMsg('Wrong');
  refresh;
end;

procedure TtstLPForm.OnLPSkip(sender : tLPMatrix);
begin
  ShowMatrix(Sender);
  OnShowSkip(sender);
end;

procedure TtstLPForm.OnShowSkip(sender : tLPMatrix);
var
  dur : tDateTime;
begin
  LPProgress.Progress := Skip;
  inc(Skip);
with sender do begin
  Dimlab.caption := inttostr(maxi)+'x'+inttostr(maxj)+'_'+IntToStr(skip);
  CastLab.Caption := FloatToStr(ToDouble(C[0]));
  GetSolution(Slice(res^,2*(ExpEd.Value+1)));
  if vc=2 then Pack2Arr(Slice(res^,2*(ExpEd.Value+1)));
  DrawPolinom(Slice(res^,ExpEd.Value+1));
end;
  dur := Now-StartTime;
  DurationLab.Caption := DurationToStr(Dur);
end;

procedure TtstLPForm.OnString(sender : tLPMatrix);
begin
  application.processmessages;
  if ShList.Count=0 then raise EAbort.Create('');
end;

procedure TtstLPForm.MinLinesClick(Sender: TObject);
var
  varMaxIdx,VarCnt,i,j,ci : integer;
  nl : TElement;
  onelimit : boolean;
  Prev : TDateTime;
begin
  Prev := Now;
  onelimit := false;
  NodesCount := 0;
  if ShList.Count=0 then OpenClick(Sender);
  if ShList.Count=0 then Exit;
  (sender as tControl).Enabled := false;
  LineCount := ShList.Count*ExpEd.Value div 2;
  ReallocMem(Lines,sizeof(tRect)*LineCount);
  try
  m := nil;
  LPProgress.Show;
  varMaxIdx := shList.Count*(shList.Count-1) div 2;
  VarCnt := shList.Count*shList.Count;
  m := tLPMatrix.CreateEx(shList.Count+VarMaxIdx*ord(onelimit),VarMaxIdx,
  varcnt,VarMaxIdx*(1+ord(onelimit)));
  m.OnChangeLeader := OnChangeLeader;
  m.OnSimplex := OnSimplex;
  m.OnSkipNode := SkipNode;
  ci := 1;
  for i := 0 to shList.Count-2 do begin
    for j := i+1 to shList.Count-1 do begin
      m.VarIdx[ci] := i*shList.Count+j;
      with m.Variables[m.VarIdx[ci]]^ do begin
        TypeLimit := Binom;
        Value := NegativeUnlimited;
      end;
      inc(ci);
    end;
  end;
  if ci<>varMaxIdx+1 then RaiseMsg('Error var ordering');
  if onelimit then
  for i := 1 to shList.Count-1 do begin
    for j := 0 to i-1 do begin
      m.VarIdx[ci] := i*shList.Count+j;
      with m.Variables[m.VarIdx[ci]]^ do begin
        TypeLimit := Binom;
        Value := NegativeUnlimited;
      end;
      inc(ci);
    end;
  end;
  nl := CreateRational(ExpEd.Value,1);
  for j := 1 to VarMaxIdx do
     m.c[j] := SizeBetween(m.VarIdx[j]);
  for i := 1 to shList.Count do begin
    m.b[i+ord(onelimit)*VarMaxIdx] := nl;
    if onelimit then begin
      m[i,i] := one;
      m.b[i] := one;
    end;
    for j := 1 to VarMaxIdx do
      if (m.VarIdx[j] div ShList.Count=(i-1))
      or (m.VarIdx[j] mod ShList.Count= (i-1)) then
        m[i+ord(onelimit)*VarMaxIdx,j] := One;
  end;
  Canvas.Pen.Color := clGreen;
  m.Solution;
//  ShowMatrix(m);
  m.UnbrokenSolution;
  Canvas.Pen.Color := clBlack;
  Repaint;
  if (Now-Prev)>=DeltaBeep then Beep;
  finally
    m.free;
   (sender as tControl).Enabled := true;
 end;
end;

procedure TtstLPForm.SolutionClick(Sender: TObject);
var
  varcnt,i,j : integer;
  Prev : TDateTime;
begin
  skip := 0;
  Prev := Now;
  LPProgress.Progress := 0;
  if shlist.Count=0 then OpenClick(Sender);
  if shlist.Count=0 then Exit;
  LPProgress.Show;
  Solution.Enabled := false;
  varcnt := (ExpEd.Value+1)*vc+2*shList.Count;
  m := tLPMatrix.Create(shList.Count,varcnt,varcnt);
  try
  LPProgress.MaxValue := varcnt;
  //  
  //     
  for i:=1 to m.MaxI do for J := 1 to ExpEd.Value+1 do begin
    m[i,vc*j] := mxel.IntPower(CreateRational(
    (shList.Objects[i-1] as tControl).left,LastX),J-1);
    if vc=2 then m[i,vc*j-1] := Negative(m[i,vc*j]);
  end;
  if vc=1 then begin
    for J := 1 to ExpEd.Value+1 do M.Variables[J].TypeLimit:=Any;
    M.AnyExists := true;
  end;
  for i := 1 to m.MaxI do
    m.B[i] := CreateRational((shList.Objects[i-1] as tControl).top,LastY);
  for j := vc*(ExpEd.Value+1)+1 to m.MaxJ do m.C[J] := One;
  for i := 1 to m.MaxI do begin
    m[i,vc*(ExpEd.Value+1)+i+m.MaxI] := One;
    m[i,vc*(ExpEd.Value+1)+i] := NegativeOne;
  end;
  M.OnSkip := OnShowSkip;
  M.OnString := OnString;
  SkipTime := Now;
  StartTime := Now;
  ReAllocMem(res,vc*(ExpEd.Value+1)*sizeof(tElement));
  Canvas.Pen.Color := clGreen;
  M.Solution;
  if not m.SolutionExists then ErrMsg('No solution');
  Canvas.Pen.Color := clBlack;
  OnShowSkip(M);
  if (Now-Prev)>=DeltaBeep then Beep;
  finally
    refresh;
    LPProgress.Hide;
    m.Free;
    Solution.Enabled := true;
  end;
end;

procedure TtstLPForm.ShMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  with Sender as tControl do
    CoordLab.Caption := IntToStr(Left)+'_'+inttostr(Top);
end;

procedure TtstLPForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  CoordLab.Caption := IntToStr(X)+' '+inttostr(Y);
  if [ssLeft,ssShift] = Shift then
    FormMouseDown(nil,mbLeft,[], X, Y);
end;

procedure TtstLPForm.FormPaint(Sender: TObject);
var
  cn : integer;
begin
  if assigned(res) then DrawPolinom(Slice(res^,ExpEd.Value+1));
  if assigned(lines) then for cn := 0 to LineCount-1 do with lines[cn] do begin
    Canvas.moveto(TopLeft.x,TopLeft.y);
    Canvas.lineto(BottomRight.x,BottomRight.y);
  end;
end;

procedure TtstLPForm.SaveasClick(Sender: TObject);
begin
  if SaveDialog.Execute then begin
    FileName := SaveDialog.FileName;
    SaveClick(self);
  end;
end;

procedure TtstLPForm.OpenClick(Sender: TObject);
var
  fop : tFileStream;
  cn : integer;
  p : tPoint;
begin
  if not OpenDialog.Execute then exit;
  FileName := OpenDialog.FileName;
  fop := tFileStream.Create(FileName,fmOpenRead);
  clearclick(self);
  try
    fop.Read(cn,sizeof(cn));
    while cn > 0 do begin
      fop.read(p,sizeof(p));
      FormMouseDown(nil,mbLeft,[], p.X, p.Y);
      dec(cn);
    end;
    fop.Read(cn,sizeof(cn));
    ExpEd.Value := cn;
  finally
    fop.Free;
  end;
end;

procedure TtstLPForm.SaveClick(Sender: TObject);
var
  fsv : tFileStream;
  cn : integer;
begin
  if FileName='' then begin
    SaveAsClick(sender);
    exit;
  end;
  fsv := tFileStream.Create(FileName,fmCreate);
  try
    cn := shlist.Count;
    fsv.write(cn,sizeof(cn));
    for cn := 0 to shlist.Count-1 do begin
      fsv.write((shlist.objects[cn] as TScrPoint).left,sizeof(integer));
      fsv.write((shlist.objects[cn] as TScrPoint).top,sizeof(integer));
    end;
    cn := ExpEd.Value;
    fsv.write(cn,sizeof(integer));
  finally
    fsv.free;
  end;
end;

procedure TtstLPForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  CanClose := IsCancel;
  if CanClose then ShList.Clear;
end;

procedure TtstLPForm.FormCreate(Sender: TObject);
begin
  ShList := tStringList.Create;
end;

end.
